/*******************************************************************************
 *	Cyrix Cx5520 integrated PLL driver
 *******************************************************************************
 *
   Copyright (C) 1998		Emmanuel Marty
   Copyright (C) 1998		Cyrix Corporation
   Copyright (C) 1998		The GGI project
 *
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *******************************************************************************
 *
 *	This code was derived from the following sources of information :
 *
 *	Cyrix MediaGX confidential databooks
 *	Cyrix MediaGX gxvideo.c example sourcecode
 *	Help and code from Sengan Baring-Gould of Cyrix Corporation
 *
 *	Many thanks to the Cyrix Corporation for providing me with
 *	documentation and official support about the GX chipset family.
 *
 *******************************************************************************
 *
 *	$Log: cx5520.c,v $
 *	Revision 1.7  1998/11/19 02:43:04  marcus
 *	License change
 *	
 *	Revision 1.6  1998/10/28 05:05:30  brian
 *	Did clock portion of cmd --> *cmd check_mode change, and some
 *	uint/sint namespace fixing (still much to do on that in clock/)
 *	
 *	Revision 1.5  1998/10/25 18:07:37  marcus
 *	More namespace cleanups
 *	
 *	Revision 1.4  1998/10/25 02:05:13  marcus
 *	Namespace cleanup to make KGI/LibGGI interaction possible
 *	
 *	Revision 1.3  1998/10/04 03:17:11  brian
 *	Added private data hook to KGI_COMMON -- clean and recompile kgicon!
 *	
 *	Revision 1.2  1998/08/19 20:47:34  marcus
 *	Changed ggi/*.h includes to kgi/*.h includes
 *	Fixed most of the remaining graphics -> accel changes
 *	
 *	Revision 1.1  1998/07/07 08:01:52  taylorj
 *	*** empty log message ***
 *	
 *	Revision 1.3  1998/03/20 01:26:04  emarty
 *	initial checkin
 *	
 */

#include <kgi/maintainers.h>
#define MAINTAINER	Emmanuel_Marty
#define DRIVER_NAME	"Cyrix Cx5520 PLL driver"
#define DRIVER_REV	"$Revision: 1.7 $"
#define DEBUG_LEVEL	0

#include <kgi/module.h>
#include <Cyrix/gx.h>
#include <Cyrix/cx5520.h>
#include "pll.h"

#define	INVALID_COMMAND	((kgiu32) -1)

static struct kgi_clock clock = {
	"Cyrix",				/* manufacturer		*/
	"Cx5520",				/* model		*/
	GM_ALL,					/* supported modes 	*/
	0,					/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },			/* maximum resolution	*/
	NULL,					/* Private data hook	*/
	CLK_TYPE_PROG,				/* clock type		*/
	{{
		25000000, 270000000,		/* clock range		*/
	}}
};

static const struct pll_limits dclk_pll = {
	14318180,			/* fref				*/
	{ 1, 1 },			/* factor		 	*/
	{ 100000000, 270000000 },	/* fvco range			*/
	{ 2, 25 },			/* post-VCO divider range	*/
	{ 2, 203 },			/* dividend range		*/
	{ 2, 203 }			/* divisor range		*/
};

/*
**	conversion table for dividend/divisor >> 1, range 1-101
**
**	this table converts from N to the value to program
**	into the chip (which directly controls the PLLs)
*/

static const kgiu16 fdiv_encode[] = {
	0x3FF,0x5FF,0x2FF,0x57F,0x2BF,0x55F,0x2AF,0x557,0x2AB,0x555,
	0x2AA,0x155,0xAA,0x455,0x22A,0x115,0x8A,0x445,0x222,0x111,
	0x88,0x444,0x622,0x311,0x588,0x6C4,0x362,0x1B1,0xD8,0x46C,
	0x636,0x31B,0x58D,0x2C6,0x163,0xB1,0x58,0x42C,0x616,0x30B,
	0x585,0x2C2,0x161,0xB0,0x458,0x62C,0x316,0x18B,0xC5,0x62,
	0x431,0x218,0x10C,0x486,0x643,0x721,0x790,0x3C8,0x1E4,0x4F2,
	0x679,0x73C,0x39E,0x1CF,0xE7,0x73,0x39,0x1C,0x40E,0x607,
	0x703,0x781,0x7C0,0x3E0,0x1F0,0x4F8,0x67C,0x33E,0x19F,0xCF,
	0x67,0x33,0x19,0xC,0x406,0x603,0x701,0x780,0x3C0,0x1E0,
	0x4F0,0x678,0x33C,0x19E,0x4CF,0x267,0x533,0x299,0x54C,
	0x6A6,0x353
};

/*
**	conversion table for post-vco divider
*/

static const kgiu16 pdiv_encode[] = {
	0x7,0x3,0x1,0x0,0x8,0x4,0xA,0x5,0x2,0x9,0xC,0x6
};

/*
**	A Cx5520 PLL command word is as follows :
**
**	[0:1]	fdiv_encode[(s->div >> 1) - 1]	(IDIV)
**	[11]	s->div & 1			(IDIV_ADDONE)
**	[12:22]	fdiv_encode[(s->mul >> 1) - 1]	(FDIV)
**	[23]	s->mul & 1			(FDIV_ADDONE)
**	[24:27]	pdiv_encode[(s->p >> 1) - 1]	(PDIV)
**	[28]	s->p & 1			(PDIV_ADDONE)
**	[29]    DCLK_VCO			(VCO enable)
*/

static kgiu32 pll_command (struct pll_state *s, struct kgi_mode *mode) {
	kgiu32 mul = s->mul, div = s->div, p = s->p;
	
	ASSERT ((mul >= 2) && (mul <= 203));
	ASSERT ((div >= 2) && (div <= 203)); 
	ASSERT ((p >= 2) && (p <= 25));

	return (fdiv_encode [(div >> 1) - 1] | (div & 1) << 11 |
		fdiv_encode [(mul >> 1) - 1] << 12 | (mul & 1) << 23 |
		pdiv_encode [(p >> 1) - 1] << 24 | (p & 1) << 28 |
		0x20000000);
}

static void pll_set (struct kgi_display *dpy, kgi_uint cmd) {
	mem_vaddr cxbase = ((struct chipset_private *)
		KGIM_PRIVATE (dpy, chipset))->gx_companion.baseptr;

	/*	reset clock		*/
	WRITE_GX32 (cxbase, CX5520_DCLK_CONTROL, cmd | 0x40000000);
	
	/*	set clock		*/
	WRITE_GX32 (cxbase, CX5520_DCLK_CONTROL, cmd);
}

/*
**	find the best parameter set to produce a certain frequency f
**	modified for Cx5520 : fout = (fref * mul/div) / p
*/

static kgi_sint pll_best_state (kgi_uint f, enum kgi_tm_cmd cmd,
	struct pll_state *s, const struct pll_limits *pll) {
	kgi_ratio	ratio;
	kgi_uint	shift;
	
	ASSERT(0 < pll->fvco.min);
	ASSERT(pll->fvco.min <= pll->fvco.max);

	ASSERT(pll->p.min <= pll->p.max);

	ASSERT(0 < pll->mul.min);
	ASSERT(pll->mul.min <= pll->mul.max);

	ASSERT(0 < pll->div.min);
	ASSERT(pll->div.min <= pll->div.max);

	/*	clear state	*/
	memset(s, 0, sizeof(*s));

	/*	determine post-VCO divider/multiplier	*/
	if ((f < pll->fvco.min) || (f > pll->fvco.max)) {

		if (f < pll->fvco.min) {
			s->p = (pll->p.min < 0) ? 0 : pll->p.min;
			s->fvco = f * s->p;

			while ((s->fvco < pll->fvco.min) &&
				(s->p < pll->p.max)) {

				s->p++;
				s->fvco = f * s->p;
			}

		} else {
			s->p = (pll->p.max > 0) ? 0 : pll->p.max;
			s->fvco = f / (- s->p);

			while ((s->fvco > pll->fvco.min) &&
				(s->p > pll->p.min)) {

				s->p--;
				s->fvco = f / (- s->p);
			}
		}

		if ((s->fvco < pll->fvco.min) || (pll->fvco.max < s->fvco) ||
			(s->p < pll->p.min) || (pll->p.max < s->p)) {

			DEBUG1("fvco = %i Hz, p = %i is out of range!", \
				s->fvco, s->p);
			return -EINVAL;
		}
	} else {
		s->fvco = f;
		s->p = 1;
	}

	if (((f * (s->p + 1)) < pll->fvco.max) && (s->p < pll->p.max)) {
		s->p++;
		s->fvco = f * s->p;
	}

	ratio.mul = s->fvco * pll->a.div;
	ratio.div = pll->fref * pll->a.mul;

	ASSERT(ratio.mul > 0);
	ASSERT(ratio.div > 0);

	/*	avoid integer overflow		*/
	shift = 0;
	while (((ratio.mul >> shift) > 0x00100000) &&
		((ratio.div >> shift) > 0x00100000)) {

		shift++;
	}

	if (shift) {

		if (cmd == CMD_LOWER) {
			ratio.mul = ratio.mul >> shift;
			ratio.div = (ratio.div - ((1 << shift) - 1)) >> shift;
		} else {
			ratio.mul = (ratio.mul - ((1 << shift) - 1)) >> shift;
			ratio.div = ratio.div >> shift;
		}
	}

	if (0 == ratio.mul  ||  0 == ratio.div) {
		DEBUG0("bad parameter set!!");
		return -EINVAL;
	}

	/*	now determine the best mul and div values.	*/
	s->div = s->mul = -1;

	if ((pll->mul.max - pll->mul.min) < (pll->div.max - pll->div.min)) {
		kgi_sint best_err = 0x7fffffff;
		kgi_sint mul = pll->mul.min;
		kgi_sint div = ((ratio.div * mul) / ratio.mul) +
			(cmd == CMD_LOWER ? 1 : 0);

		while (mul < pll->mul.max) {

			register kgi_sint err = (cmd == CMD_LOWER)
				? div * ratio.mul - mul * ratio.div
				: mul * ratio.div - div * ratio.mul;

			if (0 <= err && (err < best_err)) {

				best_err = err;
				s->mul = mul;
				s->div = div;
			}

			mul++;
			div = ((ratio.div * mul) / ratio.mul) +
				(cmd == CMD_LOWER ? 1 : 0);

			if (pll->div.max < div) {
				break;
			}
		}
		
	} else {
		kgi_sint best_err = 0x7fffffff;
		kgi_sint div = pll->div.min;
		kgi_sint mul = ((ratio.mul * div) / ratio.div) +
			(cmd == CMD_LOWER ? 0 : 1);

		while (div < pll->div.max) {

			register kgi_sint err = (cmd == CMD_LOWER)
				? div * ratio.mul - mul * ratio.div
				: mul * ratio.div - div * ratio.mul;

			if ((0 <= err) && (err < best_err)) {
				best_err = err;
				s->mul = mul;
				s->div = div;
			}

			div++;
			mul = ((ratio.mul * div) / ratio.div) +
				(cmd == CMD_LOWER ? 0 : 1);

			if (pll->mul.max < mul) {
				break;
			}
		}
	}

	if (s->mul < pll->mul.min  ||  pll->mul.max < s->mul ||
		s->div < pll->div.min || pll->div.max < s->div) {
		DEBUG1("parameters out of range");
		return -EINVAL;
	}

	s->f = pll->fref * s->mul / s->div;
	s->f = s->f * pll->a.mul / pll->a.div;
	s->f = (s->p >= 0)
		? s->f / s->p : s->f * (-s->p);

	DEBUG1("fout = %i, fvco = %i, mul = %i, div = %i, p = %i", \
		s->f, s->fvco, s->mul, s->div, s->p);
	return EOK;
}

/*
**	KGI interface
*/

kgi_sint kgim_clock_init (struct kgi_card *card, struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " rev " DRIVER_REV);
	
	card->clock = &clock;
	return EOK;
}

void kgim_clock_done (struct kgi_display *dpy) {
	dpy->card->clock = NULL;
	
	NOTICE (DRIVER_NAME " removed");
}

kgi_sint kgim_clock_command (struct kgi_graphic *state, kgi_uint cmd,
			     kgi_uint arg) {
	return -E (CLOCK, UNKNOWN);
}

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

kgi_sint kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
	enum kgi_tm_cmd *cmd) {
	struct pll_state state;
	kgi_uint newfreq = (*cmd == CMD_PROPOSE) 
		? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALID_COMMAND;
		return EOK;

	case CMD_LOWER:

		if (newfreq < MIN_FREQ) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too low",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq > MAX_FREQ) {

			newfreq = MAX_FREQ;
		}

		if (pll_best_state(newfreq, *cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_RAISE:

		if (MAX_FREQ < newfreq) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too high", \
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq < MIN_FREQ) {

			newfreq = MIN_FREQ;
		}

		if (pll_best_state(newfreq, *cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_CHECK:

		mode->clock = INVALID_COMMAND;
		if (pll_best_state(newfreq, CMD_LOWER, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->clock = pll_command(&state, mode);
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}

void kgim_clock_set_mode(struct kgi_display *dpy, struct kgi_mode *mode) {
	if (mode->clock != INVALID_COMMAND) {
		pll_set(dpy, mode->clock);
	} else {
		INTERNAL_ERROR;
	}
}
