/* $Id: setmon.c,v 1.4 1999/02/05 00:59:51 marcus Exp $
******************************************************************************

   setmon - a utility to set monitor specs for kgicon drivers

   Copyright (C) 1998 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>

#include "kgimon.h"

#define VERSION "0.3"
#define DEFAULT_DBFILE "/etc/ggi/kgicon.mon"

static const char help[] =
"Usage: %s [OPTION]... [FILE]...\n"
"Read KGI monitor specs from FILEs (the default is " DEFAULT_DBFILE
")\nand set monitor data for the specified devices.\n"
"\n"
"  -h                         display this help and exit\n"
"  -v                         output version information and exit\n\n";

static const char version[] =
"setmon v%s\n"
"(C) 1998 Marcus Sundberg [marcus@ggi-project.org]\n";

static int
parse_options(int *argc, char *argv[])
{
	int i;
	
	if (*argc < 2) {
		return 0;
	}
	
	for (i = 1; i < *argc; i++) {
		if (argv[i][0] != '-') continue;
		switch (argv[i][1]) {
		case 'v':
			printf(version, VERSION);
			return -1;
		case 'h':
			printf(help, argv[0]);
			return -1;
		default:
			fprintf(stderr, "%s: unknown option: %s\n",
				argv[0], argv[i]);
			return 1;
		}
	}
	return 0;
}

int
main(int argc, char *argv[])
{
	FILE *fp;
	char *dbfile = DEFAULT_DBFILE;
	int ret;
	int argi = 1;

	switch (ret = parse_options(&argc, argv)) {
	case -1:
		exit(0);
	case 0:
		break;
	default:
		exit(ret);
	}
	
	argc--;
	do {
		if (argc > 0) {
			dbfile = argv[argi];
		}
		if ((fp = fopen(dbfile, "r")) == NULL) {
			perror(dbfile);
			exit(1);
		}

		if ((ret = set_monfromfile(fp)) != 0) {
			switch (ret) {
			case END_OF_FILE:
				fprintf(stderr, "end-of-file encountered in %s\n",
					dbfile);
				break;
			case PARSE_ERROR:
				fprintf(stderr, "parse error in %s\n",
					dbfile);
				break;
			}
		}
		argc--;
		argi++;
	} while (argc > 0);
	
	return 0;
}
